var express = require('express')
var FireRuntime = require('fire').Runtime
var sys = require('util')
var FinishError = require('./FinishError.js')

function WebServer() {
	
}

WebServer.prototype.setRuntime = function(runtime) {
	this.runtime = runtime
	var app = this.expressApp = express.createServer();
	this.expressApp.runtime = this.runtime
	this.runtime.web = this
	this.expressApp.configure(function(){
	  app.set('views', this.runtime.pathFromBaseDir('views'));
	  app.set('view engine', 'jade');
	  app.use(express.bodyParser());
	  app.use(express.methodOverride());
	  app.use(app.router);
	  app.use(express.static(this.runtime.pathFromBaseDir('public')));
	});
	this.config = this.runtime.getModuleConfiguration('fire-web')
}

WebServer.prototype._finishWithProtocol = function(expressVar, expressionDef) {
	var self = this
	var responseProtocolName = expressionDef["Web.Response.Protocol"] || this.config.defaultResponseProtocol
	var expressionDef = this.runtime.getExpressionDefinition(responseProtocolName)
	
	var exp = new(expressionDef.implementation)
	exp.runtime = self.runtime
	exp.resultCallback = function(res) {
		// Request Finished
	}
	exp.loopCallback = function() {
		throw "WebServer catched a loopCallback command on response protocol"
	};
	exp.errorCallback = function(err) {
		if(err.error instanceof FinishError) {
			self._finishWithProtocol(expressVar, expressionDef)
		} else {
			showError(err, expressReq, expressRes)
		}
	}
	exp.ensureInitialized()
	exp.setVar("express", expressVar)
	exp.run()
}

WebServer.prototype._prepareWebApp = function() {
	var definitions = this.runtime.getWellKnownExpressions()
	var self = this
	definitions.names().forEach(function(moduleName) {
		var expressionDef = definitions[moduleName]
		
		var expressionMethod = null
		var expressionRoute = null
		
		if(expressionDef['Web.Endpoint']) {
			var endPoint = expressionDef['Web.Endpoint']
			var space = endPoint.indexOf(' ')
			expressionMethod = endPoint.slice(0, space)
			expressionRoute = endPoint.slice(space +1)
		}
		
		if(expressionRoute && expressionMethod) {
			// Yay, we found an expression exposed for web requests.
			var httpMethod = expressionMethod.toLowerCase()
			self.expressApp[httpMethod](expressionRoute, function(expressReq, expressRes, expressNext) {
				var expressVar = {
					req: expressReq,
					res: expressRes
				}
				var exp = new(expressionDef.implementation)
				exp.runtime = self.runtime
				exp.resultCallback = function(result) {
					expressRes.webResponse = result
					self._finishWithProtocol(expressVar, expressionDef)
				}
				exp.loopCallback = function() {
					throw "WebServer catched a loopCallback command"
				};
				exp.errorCallback = function(err) {
					if(err.error instanceof FinishError) {
						self._finishWithProtocol(expressVar, expressionDef)
					} else {
						showError(err, expressReq, expressRes)
					}
				}
				exp.ensureInitialized()
				exp.setVar("express", expressVar)
				exp.run()
			})
		}
	});
}

function showError(err, req, res) {
	console.error("Server Internal Error:", err, " : ",err.stack, "\n")
	req.statusCode = 500
	req.next("Something went wrong")
}

WebServer.prototype.boot = function() {
	var self = this
	// Extract all expression ready for web publishing and make them routes in express.js
	
	this._prepareWebApp()
	
	var port = process.env.VMC_APP_PORT // cloudfoundry
	|| process.env.PORT // heroku
	|| this.config.port // local & custom environments
	
	sys.print("Booting...")
	this.expressApp.on("listening", function() {
		self.runtime.events.emit('web')
		sys.print("fire.js WebServer running at port " + self.config.port + "\n")
	})
	this.expressApp.listen(port)
}

module.exports.WebServer = WebServer
